from jinja2 import Environment, BaseLoader
from girder_client import HttpError
import re
import avogadro

from ._girder import GirderClient
from ._molecule import Molecule
from ._calculation import (
    GirderMolecule, CalculationResult, AttributeInterceptor,
    _fetch_calculation, _fetch_or_submit_calculations, _calculation_result
)
from ._data import CjsonProvider, AvogadroProvider
from ._utils import fetch_or_create_queue

_inchi_key_regex = re.compile("^([0-9A-Z\-]+)$")

def _is_inchi_key(identifier):
    return len(identifier) == 27 and identifier[25] == '-' and \
        _inchi_key_regex.match(identifier)

def _find_molecule(identifier=None, inchi=None, smiles=None):
    if inchi:
        return _find_molecule_by_inchi(inchi)

    if smiles:
        return _find_molecule_by_smiles(smiles)

    # InChiKey?
    if _is_inchi_key(identifier):
        return _find_molecule_by_inchikey(identifier)

    # Finally, if identifier is something else (a name), try cactus
    if identifier:
        return _find_molecule_using_cactus(identifier)

    return None

def _find_molecule_by_inchi(inchi):
    params = {
        'inchi': inchi
    }
    return _find_molecule_using_girder(params)

def _find_molecule_by_smiles(smiles):
    params = {
        'smiles': smiles
    }
    return _find_molecule_using_girder(params)

def _find_molecule_by_inchikey(inchikey):
    try:
        return GirderClient().get('molecules/inchikey/%s' % inchikey)
    except HttpError as ex:
        if ex.status == 404:
            return _find_molecule_using_cactus(inchikey)
        else:
            raise

def _find_molecule_using_cactus(identifier):
    params = {
        'cactus': identifier
    }
    res = GirderClient().get('molecules/search', parameters=params)
    # Just pick the first
    if 'results' in res and len(res['results']) > 0:
        return res['results'][0]
    else:
        return None

def _find_molecule_using_girder(params):
    res = GirderClient().get('molecules', parameters=params)
    if 'results' in res and len(res['results']) > 0:
        return res['results'][0]
    else:
        return None

def _calculation_monitor(taskflow_ids):
    try:
        from ._notebook import CalculationMonitor
        table = CalculationMonitor({
            'taskFlowIds': taskflow_ids,
            'girderToken': GirderClient().token,
            'girderApiUrl': GirderClient().api_url
        })
    except ImportError:
        # Outside notebook just print message
        table = 'Pending calculations .... '

    return table

def import_structure(smiles=None, inchi=None, cjson=None, gen3d=True,
                     params=None, gen3d_forcefield='mmff94', gen3d_steps=100):
    # If the smiles begins with 'InChI=', then it is actually an inchi instead
    if smiles and smiles.startswith('InChI='):
        inchi = smiles
        smiles = None

    if params is None:
        params = {}
    if smiles:
        params['smiles'] = smiles
    elif inchi:
        params['inchi'] = inchi
    elif cjson:
        params['cjson'] = cjson
    else:
        raise Exception('SMILES, InChI, or CJson must be provided')

    params['generate3D'] = gen3d
    params['gen3dForcefield'] = gen3d_forcefield
    params['gen3dSteps'] = gen3d_steps
    molecule = GirderClient().post('molecules', json=params)

    if not molecule:
        raise Exception('Molecule could not be imported with params', params)

    return GirderMolecule(molecule['_id'], molecule.get('cjson'))

def find_molecule(identifier=None, inchi=None, smiles=None):
    molecule = _find_molecule(identifier, inchi, smiles)
    if molecule is None:
        raise Exception('Unable to find a molecule with the provided identifiers.')
    return GirderMolecule(molecule['_id'], molecule.get('cjson'))

def find_calculation(molecule, image_name=None, input_parameters=None, input_geometry=None):
    calculation = _fetch_calculation(molecule._id, image_name, input_parameters, input_geometry)
    if calculation is None:
        raise Exception('Unable to find a matching calculation in the database')
    return CalculationResult(calculation['_id'])

def find_structure(identifier=None, image_name=None, input_parameters=None, input_geometry=None, inchi=None, smiles=None):
    molecule = find_molecule(identifier, inchi, smiles)

    # If we have been provided basis, theory or functional it means the user is
    # looking for a calculation, otherwise they're looking for a molecule
    is_calc_query = (image_name is not None and input_parameters is not None)

    if is_calc_query:
        return find_calculation(molecule, image_name, input_parameters, input_geometry)
    else:
        return molecule

def load(data):
    if isinstance(data, dict):
        provider = CjsonProvider(data)
    elif isinstance(data, avogadro.core.Molecule):
        provider = AvogadroProvider(data)
    else:
        raise TypeError("Load accepts either a cjson dict, or an avogadro.core.Molecule")
    return Molecule(provider)

def monitor(results):
    taskflow_ids = []

    for result in results:
        if hasattr(result, '_properties'):
            props = result._properties
            if isinstance(props, AttributeInterceptor):
                props = props.unwrap()
            if isinstance(props, dict):
                taskflow_id = props.get('taskFlowId')
                if taskflow_id is not None:
                    taskflow_ids.append(taskflow_id)

    return _calculation_monitor(taskflow_ids)

def queue():
    if GirderClient().host is None:
        import warnings
        warnings.warn("Cannot displaying pending calculations, the notebook is not running in a Girder environment")
        return

    queue = fetch_or_create_queue(GirderClient())
    running = []
    pending = []

    for taskflow_id, status in queue['taskflows'].items():
        if status == 'pending':
            pending.append(taskflow_id)
        elif status == 'running':
            running.append(taskflow_id)

    taskflow_ids = running + pending

    return _calculation_monitor(taskflow_ids)

def find_spectra(identifier, stype='IR', source='NIST'):
    """Find spectra in source database

    Parameters
    ----------
    identifier : str
        Inchi string.
    stype : str
        Type of spectrum to query.
    source : str
        Database to query. Supported are: 'NIST'
    """
    source = source.lower()

    inchi = GirderClient().get('molecules/{}'.format(identifier))["inchi"]

    params = {
            'inchi' : inchi,
            'spectrum_type' : stype,
            'source' : source
    }

    spectrum = GirderClient().get('experiments', parameters=params)

    if spectrum is not None:
        frequencies = [float(w) for w in spectrum['x'][1:-1].split()]
        intensities = [float(i) for i in spectrum['y'][1:-1].split()]
        max_intensity = max(intensities)
        intensities = [i / max_intensity for i in intensities]
        spectrum = {'intensities': intensities, 'frequencies': frequencies}

    return spectrum

def run_calculations(girder_molecules, image_name, input_parameters,
                     input_geometries=None, run_parameters=None,
                     force=False):
    """Run multiple calculations in one taskflow

    This will search for each calculation to see if it has already been
    done. If it has been done, it will not run it unless force is true.
    It will then submit all calculations that need to be submitted.

    Parameters
    ----------
    girder_molecules : list of GirderMolecule
        The list of molecules for which to run calculations
    image_name : str
        The name of the image to run the calculations on
    input_paramters : dict
        The input parameters for the taskflow
    input_geometries: list of str
        The input geometries of the molecules
    run_paramters : dict
        The run parameters for the taskflow
    force : bool
        Force all of the calculations to be performed, even if they
        have already been run once.
    """
    molecule_ids = [x._id for x in girder_molecules]
    calculations = _fetch_or_submit_calculations(molecule_ids, image_name,
                                                 input_parameters,
                                                 input_geometries,
                                                 run_parameters, force)

    results = []
    for c, m in zip(calculations, molecule_ids):
        results.append(_calculation_result(c, m))

    return results



from abc import ABC, abstractmethod
import json
import collections
import avogadro

from ._girder import GirderClient
from ._utils import calculate_mo

from girder_client import HttpError

class DataProvider(ABC):
    @property
    @abstractmethod
    def cjson(self):
        pass

    @property
    @abstractmethod
    def vibrations(self):
        pass

    @abstractmethod
    def load_orbital(self, mo):
        pass

    @property
    @abstractmethod
    def url(self):
        pass

class CachedDataProvider(DataProvider):
    MAX_CACHED = 5

    def __init__(self):
        self._cached_volumes = collections.OrderedDict()

    def _get_cached_volume(self, mo):
        if mo in self._cached_volumes:
            return self._cached_volumes[mo]
        return None

    def _set_cached_volume(self, mo, cube):
        keys = self._cached_volumes.keys()
        if len(keys) >= self.MAX_CACHED:
            del self._cached_volumes[next(iter(keys))]
        self._cached_volumes[mo] = cube

class CjsonProvider(CachedDataProvider):
    def __init__(self, cjson):
        super(CjsonProvider, self).__init__()
        self._cjson_ = cjson

    @property
    def cjson(self):
        return self._cjson_

    @property
    def vibrations(self):
        if 'vibrations' in self.cjson:
            return self.cjson['vibrations']
        else:
            return {'modes': [], 'intensities': [], 'frequencies': []}

    def load_orbital(self, mo):
        cube = super(CjsonProvider, self)._get_cached_volume(mo)
        if cube is None:
            cube = calculate_mo(self.cjson, mo)
            super(CjsonProvider, self)._set_cached_volume(mo, cube)
        cjson = self.cjson
        cjson['cube'] = cube

    @property
    def url(self):
        return None

class MoleculeProvider(CjsonProvider):
    def __init__(self, cjson, molecule_id):
        super(MoleculeProvider, self).__init__(cjson)
        self._id = molecule_id
        self._svg_ = None

    @property
    def cjson(self):
        if self._cjson_ is None:
            # Try to update the cjson
            self._cjson_ = GirderClient().get('molecules/%s/cjson' % self._id)

        return self._cjson_

    @property
    def svg(self):
        if self._svg_ is None:
            resp = GirderClient().get('molecules/%s/svg' % self._id,
                                      jsonResp=False)
            self._svg_ = resp.content.decode('utf-8')

        return self._svg_

    def geometry_cjson(self, geometry_id=None):
        if geometry_id is None:
            # Just return the molecule's default cjson
            return self.cjson

        try:
            resp = GirderClient().get(
                'molecules/%s/geometries/%s' % (self._id, geometry_id))
        except HttpError as e:
            if ('Invalid ObjectId' in e.responseText or
                    'Geometry not found' in e.responseText):
                print('Geometry ID not found')
                return None

            raise

        return resp.get('cjson')

    @property
    def geometries(self):
        resp = GirderClient().get('molecules/%s/geometries' % self._id)
        return resp.get('results', [])

    @property
    def url(self):
        return '%s/molecules/%s' % (GirderClient().app_url.rstrip('/'), self._id)

class CalculationProvider(CachedDataProvider):
    def __init__(self, calculation_id, molecule_id):
        super(CalculationProvider, self).__init__()
        self._id = calculation_id
        self._molecule_id = molecule_id
        self._cjson_ = None
        self._vibrational_modes_ = None

    @property
    def cjson(self):
        if self._cjson_ is None:
            self._cjson_ = GirderClient().get('calculations/%s/cjson' % self._id)

        return self._cjson_

    @property
    def vibrations(self):
        if self._vibrational_modes_ is None:
            self._vibrational_modes_ = GirderClient().get('calculations/%s/vibrationalmodes' % self._id)

        return self._vibrational_modes_

    def load_orbital(self, mo):
        cube = super(CalculationProvider, self)._get_cached_volume(mo)
        if cube is None:
            cube = GirderClient().get('calculations/%s/cube/%s' % (self._id, mo))['cube']
            super(CalculationProvider, self)._set_cached_volume(mo, cube)
        cjson = self.cjson
        cjson['cube'] = cube

    @property
    def url(self):
        return '%s/calculations/%s' % (GirderClient().app_url.rstrip('/'), self._id)

class AvogadroProvider(CjsonProvider):
    def __init__(self, molecule):
        super(AvogadroProvider, self).__init__(None)
        self._molecule = molecule

    @property
    def cjson(self):
        if self._cjson_ is None:
            conv = avogadro.io.FileFormatManager()
            cjson_str = conv.write_string(self._molecule, 'cjson')
            self._cjson_ = json.loads(cjson_str)
        return self._cjson_



from .utils import (
    _cclib_to_cjson_basis,
    _cclib_to_cjson_vibdisps,
)
from .base import BaseReader
from .constants import EV_TO_J_MOL


class OrcaReader(BaseReader):
    """A class to parse orca output files and dump a chemical json file"""

    def read(self):
        """Read orca output file"""
        import cclib
        from pybel import readfile

        molecule = next(readfile("orca", self._file))

        atom_numbers = []
        coordinates = []

        for atom in molecule:
            atom_number = atom.atomicnum
            atom_numbers.append(atom_number)
            for coord in atom.coords:
                coordinates.append(coord)

        molecular_mass = molecule.exactmass

        cjson = {
            "chemical json": 0,
            "atoms": {
                "elements": {"number": atom_numbers},
                "coords": {"3d": coordinates},
            },
            "properties": {"molecular mass": molecular_mass},
        }

        data = cclib.io.ccread(self._file)

        # Add calculated properties
        if hasattr(data, "scfenergies"):
            if len(data.scfenergies) > 0:
                energy = data.scfenergies[-1] * EV_TO_J_MOL
                cjson.setdefault("properties", {})["totalEnergy"] = energy

        if hasattr(data, "gbasis"):
            basis = _cclib_to_cjson_basis(data.gbasis)
            cjson["basisSet"] = basis

        if hasattr(data, "vibfreqs"):
            vibfreqs = list(data.vibfreqs)
            cjson.setdefault("vibrations", {})["frequencies"] = vibfreqs

        if hasattr(data, "vibdisps"):
            vibdisps = _cclib_to_cjson_vibdisps(data.vibdisps)
            cjson.setdefault("vibrations", {})["eigenVectors"] = vibdisps

        # Add a placeholder intensities array
        if "vibrations" in cjson and "frequencies" in cjson["vibrations"]:
            if "intensities" not in cjson["vibrations"]:
                cjson["vibrations"]["intensities"] = [
                    1 for i in range(len(cjson["vibrations"]["frequencies"]))
                ]
            if "modes" not in cjson["vibrations"]:
                cjson["vibrations"]["modes"] = [
                    i + 1 for i in range(len(cjson["vibrations"]["frequencies"]))
                ]

        return cjson



from ._visualization import (
    Structure, Orbitals, Properties, Vibrations, Geometries
)

class Molecule(object):
    def __init__(self, provider):
        self._provider = provider
        self._visualizations = {
            'structure': None,
            'orbitals': None,
            'properties': None,
            'vibrations': None,
            'geometries': None
        }

    @property
    def structure(self):
        if self._visualizations['structure'] is None:
            self._visualizations['structure'] = Structure(self._provider)
        return self._visualizations['structure']

    @property
    def orbitals(self):
        if self._visualizations['orbitals'] is None:
            self._visualizations['orbitals'] = Orbitals(self._provider)
        return self._visualizations['orbitals']

    @property
    def properties(self):
        if self._visualizations['properties'] is None:
            self._visualizations['properties'] = Properties(self._provider)
        return self._visualizations['properties']

    @property
    def vibrations(self):
        if self._visualizations['vibrations'] is None:
            self._visualizations['vibrations'] = Vibrations(self._provider)
        return self._visualizations['vibrations']

    @property
    def geometries(self):
        if self._visualizations['geometries'] is None:
            self._visualizations['geometries'] = Geometries(self._provider)
        return self._visualizations['geometries']



import os
from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

def prerelease_local_scheme(version):
    """Return local scheme version unless building on master in CircleCI.
    This function returns the local scheme version number
    (e.g. 0.0.0.dev<N>+g<HASH>) unless building on CircleCI for a
    pre-release in which case it ignores the hash and produces a
    PEP440 compliant pre-release version number (e.g. 0.0.0.dev<N>).
    """

    from setuptools_scm.version import get_local_node_and_date

    if 'CIRCLE_BRANCH' in os.environ and \
       os.environ.get('CIRCLE_BRANCH') == 'master':
        return ''
    else:
        return get_local_node_and_date(version)

setup(
    name='openchemistry',
    use_scm_version={'local_scheme': prerelease_local_scheme},
    setup_requires=['setuptools_scm'],
    description='',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/OpenChemistry/openchemistrypy',
    author='Kitware Inc',
    license='BSD 3-Clause',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.6'
    ],

    keywords='',

    packages=find_packages(exclude='taskflows'),

    install_requires=[
        'girder_client>=2.3.0',
        'jinja2',
        'jsonpath-rw',
        'avogadro>=1.92.1',
        'IPython',
        'ipykernel'
    ],

    extras_require={

    }
)



from ._utils import hash_object, camel_to_space, parse_image_name
from .io import Psi4Reader, NWChemJsonReader, CjsonReader, OrcaReader
from .api import (
    load, find_structure, find_calculation, find_molecule, monitor, queue,
    find_spectra, import_structure, run_calculations
)
